import asyncio
import json
import os
import sys
from typing import Any, Dict, Union

from .exceptions import TgEraserException

TG_API_ID = os.environ.get("TG_API_ID")
TG_API_HASH = os.environ.get("TG_API_HASH")


async def async_input(prompt: str) -> str:
    """Non-blocking input"""
    try:
        print(prompt, end="", flush=True)
        return (
            await asyncio.get_running_loop().run_in_executor(None, sys.stdin.readline)
        ).rstrip()
    except Exception as e:
        raise TgEraserException(f"Error in async_input: {e}")


def sprint(string: str, *args: Any, **kwargs: Any) -> None:
    """Safe Print"""
    try:
        print(string, *args, **kwargs)
    except UnicodeEncodeError:
        string = string.encode("utf-8", errors="ignore").decode(
            "ascii", errors="ignore"
        )
        print(string, *args, **kwargs)


def print_header(title: str) -> None:
    """Print titles to the console nicely"""
    sprint(f"\n=={'=' * len(title)}==")
    sprint(f"= {title} =")
    sprint(f"=={'=' * len(title)}==")


def cast_to_int(num: str, name: str) -> int:
    """Check if a string represents an int"""
    try:
        return int(num)
    except ValueError as err:
        raise TgEraserException(f"Error: '{name}' should be an integer.") from err


async def get_credentials(args: Dict[str, Any]) -> Dict[str, Union[str, int]]:
    """
    Get credentials from file or from user input
    """
    path_to_creds_dir = os.path.abspath(os.path.expanduser(args["--directory"]))
    path_to_creds_file = path_to_creds_dir + "/credentials.json"
    os.makedirs(path_to_creds_dir, exist_ok=True)

    creds = {}

    try:
        if os.path.exists(path_to_creds_file):
            with open(path_to_creds_file, "r") as file:
                creds = json.load(file)
        elif TG_API_ID and TG_API_HASH:
            creds["api_id"] = cast_to_int(TG_API_ID, "api_id")
            creds["api_hash"] = TG_API_HASH
        else:
            creds["api_id"] = cast_to_int(
                await async_input("Enter your API ID: "), "api_id"
            )
            creds["api_hash"] = await async_input("Enter your API hash: ")

            save_to_file = await async_input(
                "Do you want to save your credentials to file? [y/n]: "
            )
            if save_to_file.lower() == "y":
                with open(path_to_creds_file, "w") as file:
                    json.dump(creds, file)
                sprint(f"Credentials saved to '{path_to_creds_file}' file.")

        creds["session_name"] = (
            path_to_creds_dir
            + "/"
            + (args["<session_name>"] or await choose_session(path_to_creds_dir))
        )
    except Exception as e:
        raise TgEraserException(f"Error in get_credentials: {e}")

    return creds


async def choose_session(directory: str) -> str:
    """Choose a session or enter a new one"""
    sessions = list_sessions(directory)

    if sessions:
        print_header("Choose one of the following sessions:")
        for i, session in enumerate(sessions):
            print(f"{i + 1}. {session}")
        session_number = cast_to_int(
            await async_input("Enter session number: "), "session number"
        )
        return sessions[session_number - 1]
    else:
        return await async_input("Enter session name: ")


def list_sessions(directory: str) -> list[str]:
    """Return file names with '.session' extension in specified directory"""
    return [
        os.path.splitext(file)[0]
        for file in os.listdir(directory)
        if file.endswith(".session")
    ]



"""
Tool deletes all your messages from chat/channel/conversation on Telegram.

Usage:
    tgeraser [(session <session_name>) --entity-type TYPE -l NUM [-d PATH] -p PEER_ID] | [-k]
    tgeraser session <session_name> -p PEER_ID -t STRING
    tgeraser session <session_name> -w [--entity-type TYPE]
    tgeraser -h | --help
    tgeraser --version

Options:
    -d --directory PATH         Specify a directory where your sessions are stored. [default: ~/.tgeraser/]
    -w --wipe-everything        Delete all messages from all entities of a certain type that you have in your dialog list.
    --entity-type TYPE          Available types: any, chat, channel, user. [default: chat]
    -p --peers PEER_ID          Specify certain peers by comma (chat/channel/user).
    -l --limit NUM              Show a specified number of recent chats.
    -t --time-period STRING     Specify a period for an infinite loop to run messages deletion every X seconds/minutes/hours/days/weeks.
                                Example: --time-period "3*days" OR --time-period "5*seconds"
    -k --kill                   Terminate existing background TgEraser processes (only for Unix-like OS).
    -h --help                   Show this screen.
    --version                   Show version.

"""  # pylint: disable=line-too-long

import asyncio
import os
import signal
import subprocess
import sys
import time

from docopt import docopt

from .__version__ import VERSION
from .eraser import Eraser
from .exceptions import TgEraserException
from .utils import cast_to_int, get_credentials


async def main() -> None:
    """
    Entry function
    """
    arguments = docopt(__doc__, version=VERSION)
    if arguments["--limit"]:
        arguments["--limit"] = cast_to_int(arguments["--limit"], "limit")
    if arguments["--time-period"]:
        periods = {
            "seconds": 1,
            "minutes": 60,
            "hours": 3600,
            "days": 86400,
            "weeks": 604800,
        }
        period = arguments["--time-period"].split("*")
        if period[1] not in periods:
            raise TgEraserException("Time period is specified incorrectly.")

    if arguments["--kill"]:
        if os.name != "posix":
            raise TgEraserException("You can't use '--kill' option on Windows.")
        cmd = subprocess.Popen(["ps", "-A"], stdout=subprocess.PIPE)
        out = cmd.communicate()[0]
        current_pid = os.getpid()
        for line in out.splitlines():
            if b"tgeraser" in line:
                pid = int(line.split(None, 1)[0])
                if current_pid != pid:
                    os.kill(pid, signal.SIGKILL)
                    print("Process " + str(pid) + " successfully killed.")
        sys.exit(0)

    try:
        credentials = await get_credentials(arguments)

        kwargs = {
            **credentials,
            "peers": arguments["--peers"],
            "limit": arguments["--limit"],
            "wipe_everything": arguments["--wipe-everything"],
            "entity_type": arguments["--entity-type"],
        }

        client = Eraser(**credentials)
        while True:
            await client.init(**kwargs)
            await client.run()
            if arguments["--time-period"]:
                print(
                    "\n({0})\tNext erasing will be in {1} {2}.".format(
                        time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime()),
                        period[0],
                        period[1],
                    )
                )
                await asyncio.sleep(int(period[0]) * periods[period[1]])
            else:
                break
        await client.disconnect()
    except asyncio.CancelledError as err:
        print("\n\nCancelled by user.\n", file=sys.stderr)
    except Exception as err:
        raise TgEraserException(err) from err


def entry() -> None:
    """
    Entry point function
    """
    asyncio.run(main())


if __name__ == "__main__":
    entry()
